﻿Option Explicit On
Option Strict Off

Imports Game_PluginAPI
Imports System.Math
Imports System.Runtime.InteropServices

Public Class GamePlugin
    Implements IPlugin_Game
    '//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
    '///                            SimTools Plugin - Edit the Setting below to provide support for your favorite game!                             ///
    '//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
    '//////////////////////////////////////////////// 
    '/// Per Game Settings - Change for Each Game ///
    '////////////////////////////////////////////////
    Private Const _PluginAuthorsName As String = "value1, yobuddy"
    Private Const _GameName As String = "F1_2022" 'GameName (Must Be Unique!) - the displayed Name.
    Private Const _ProcessName As String = "F1_22" 'Process_Name without the (".exe") for this game
    Private Const _Port As String = "20777"  'Your Sending/Recieving UDP Port for this game
    Private Const _RequiresPatchingPath As Boolean = True 'do we need the game exe path for patching? (must be True if _RequiresSecondCheck = True)
    Private Const _RequiresSecondCheck As Boolean = False 'Used when games have the same _ProcessName. (all plugins with the same _ProcessName must use this!)
    Private Const _PluginOptions As String = "" 'Reserved For Future Use - No Change Needed 
    '////////////////////////////////////////////////
    '///           Memory Map Variables           ///
    '////////////////////////////////////////////////
    Private Const _Enable_MemoryMap As Boolean = False 'Is a MemoryMap file Required for this game?
    Private Const _MMF_Name As String = "NULL" 'Set if using a Memory Map File - EXAMPLE("$gtr2$")
    '////////////////////////////////////////////////
    '///           MemoryHook Variables           ///
    '////////////////////////////////////////////////   
    Private Const _Enable_MemoryHook As Boolean = False 'Is a Memory Hook Required for this game? 
    Private Const _MemHook_Roll As UInteger = 0 'Not Used = 0
    Private Const _MemHook_Pitch As UInteger = 0
    Private Const _MemHook_Heave As UInteger = 0
    Private Const _MemHook_Yaw As UInteger = 0
    Private Const _MemHook_Sway As UInteger = 0
    Private Const _MemHook_Surge As UInteger = 0
    Private Const _MemHook_Extra1 As UInteger = 0
    Private Const _MemHook_Extra2 As UInteger = 0
    Private Const _MemHook_Extra3 As UInteger = 0
    '////////////////////////////////////////////////
    '///    DOFs Used for Output for this Game    ///
    '////////////////////////////////////////////////
    Private Const _DOF_Support_Roll As Boolean = True
    Private Const _DOF_Support_Pitch As Boolean = True
    Private Const _DOF_Support_Heave As Boolean = True
    Private Const _DOF_Support_Yaw As Boolean = True
    Private Const _DOF_Support_Sway As Boolean = True
    Private Const _DOF_Support_Surge As Boolean = True
    Private Const _DOF_Support_Extra1 As String = "Traction Loss" 'Blank = False
    Private Const _DOF_Support_Extra2 As String = "" '"" = Not Used
    Private Const _DOF_Support_Extra3 As String = "" 'ADD THE FORCE NAME HERE
    '/////////////////////////////////////////////////
    '///       GameDash - Dash Board Support       ///
    '/////////////////////////////////////////////////
    Private Const _Enable_DashBoard As Boolean = True  'Enable the DashBoard Output System?
    'EXAMPLES OF DASH OUTPUT (all variables are strings) - 20 DASH OUTPUTS MAX!!!
    'Use the Variables with either Process_PacketRecieved, Process_MemoryHook or Process_MemoryMap.
    'Variable = (*Action, Value) as String - (The *Action variable can be anything)
    'Basic Dash Support should include at the minimum -  SPEED, RPM and GEAR
    'Dash_1_Output = "Speed," & OUTPUT_VALUE_HERE.ToString
    'Dash_2_Output = "Rpm," & OUTPUT_VALUE_HERE.ToString
    'Dash_2_Output = "Gear," & OUTPUT_VALUE_HERE.ToString
    '...
    'Dash_20_Output = "Engine Temp," & OUTPUT_VALUE_HERE.ToString
    '/////////////////////////////////////////////////
    '///           GameVibe Support              ///
    '/////////////////////////////////////////////////
    Private Const _Enable_GameVibe As Boolean = True 'Enable the GameVibe Output System?
    'EXAMPLES OF GameVibe OUTPUT (all variables are strings) - 9 GameVibe OUTPUTS MAX!!!
    'Use the Variables with either Process_PacketRecieved, Process_MemoryHook or Process_MemoryMap.
    'Variable = (*Action, Value) as String - (The *Action" variable can ONLY be: Rpm, Heave, Sway, Surge, FR, FL, RR, RL, Turbo)
    'Basic GameVibe Support should include at the minimum -  Rpm
    'Vibe_1_Output = "Rpm," & OUTPUT_VALUE_HERE.ToString
    'Vibe_2_Output = "Heave," & OUTPUT_VALUE_HERE.ToString
    'Vibe_3_Output = "Sway," & OUTPUT_VALUE_HERE.ToString
    'Vibe_4_Output = "Surge," & OUTPUT_VALUE_HERE.ToString
    '...
    'Vibe_9_Output = "Gear Shift," & OUTPUT_VALUE_HERE.ToString
    '/////////////////////////////////////////////////

    'Used by GameManager when the Game Starts.
    Public Sub GameStart() Implements IPlugin_Game.GameStart
        'DO SOMETHING HERE AT GAME START!
    End Sub

    'Used by GameManager when the Game Stops.
    Public Sub GameStop() Implements IPlugin_Game.GameEnd
        'DO SOMETHING HERE AT GAME STOP!
    End Sub

    'Used by GameManager to Process a MemoryHook.
    Public Sub Process_MemoryHook() Implements IPlugin_Game.Process_MemoryHook
        'DO SOMETHING HERE AT GAME START!        
    End Sub

    'Used by GameManager to Process a MemoryMap.
    Public Sub Process_MemoryMap() Implements IPlugin_Game.Process_MemoryMap
        'DO SOMETHING HERE AT GAME START!
    End Sub

    'Used by GameEngine to Process Incoming UDP Packets.
    Public Sub Process_PacketRecieved(Text As String) Implements IPlugin_Game.Process_PacketRecieved
        'Unit conversions
        Dim radtodeg As Single = CSng(45.0 / Atan(1))
        Dim gtompss As Single = 9.80665 'g to m/s²

        Dim ByteArray() As Byte = System.Text.Encoding.Default.GetBytes(Text)
        If ByteArray(0) = 226 And ByteArray(1) = 7 Then
            If ByteArray(3) = 0 Then
                Dim CodemastersData As New PacketMotionData
                Dim gch As GCHandle = GCHandle.Alloc(ByteArray, GCHandleType.Pinned)
                CodemastersData = Marshal.PtrToStructure(gch.AddrOfPinnedObject, CodemastersData.GetType)
                gch.Free()
                With CodemastersData
                    Yaw_Output = .m_carMotionData(.m_header.m_playerCarIndex).m_yaw * radtodeg
                    Roll_Output = .m_carMotionData(.m_header.m_playerCarIndex).m_roll * radtodeg
                    Pitch_Output = .m_carMotionData(.m_header.m_playerCarIndex).m_pitch * radtodeg
                    Surge_Output = .m_carMotionData(.m_header.m_playerCarIndex).m_gForceLongitudinal
                    Sway_Output = .m_carMotionData(.m_header.m_playerCarIndex).m_gForceLateral
                    Heave_Output = .m_carMotionData(.m_header.m_playerCarIndex).m_gForceVertical
                    Extra1_Output = .m_localVelocityX

                    Vibe_3_Output = "Collision L/R," & Sway_Output.ToString
                    Vibe_4_Output = "Collision F/B," & Surge_Output.ToString
                    Vibe_5_Output = "Road Detail," & Heave_Output.ToString
                    Vibe_6_Output = "Chassis," & .m_suspensionPosition(0) & ":" & .m_suspensionPosition(1) & ":" & .m_suspensionPosition(2) & ":" & .m_suspensionPosition(3)
                End With

            ElseIf ByteArray(3) = 1 Then
                Dim CodemastersData As New PacketSessionData
                Dim gch As GCHandle = GCHandle.Alloc(ByteArray, GCHandleType.Pinned)
                CodemastersData = Marshal.PtrToStructure(gch.AddrOfPinnedObject, CodemastersData.GetType)
                gch.Free()
                With CodemastersData
                    Dash_12_Output = "Track," & .m_trackId
                End With

            ElseIf ByteArray(3) = 2 Then
                Dim CodemastersData As New PacketLapData
                Dim gch As GCHandle = GCHandle.Alloc(ByteArray, GCHandleType.Pinned)
                CodemastersData = Marshal.PtrToStructure(gch.AddrOfPinnedObject, CodemastersData.GetType)
                gch.Free()
                With CodemastersData
                    Dash_11_Output = "In pits," & .m_lapData(.m_header.m_playerCarIndex).m_pitStatus
                End With

            ElseIf ByteArray(3) = 3 Then
                Dim CodemastersData As New PacketEventData
                Dim gch As GCHandle = GCHandle.Alloc(ByteArray, GCHandleType.Pinned)
                CodemastersData = Marshal.PtrToStructure(gch.AddrOfPinnedObject, CodemastersData.GetType)
                gch.Free()
                With CodemastersData
                    'Dash_1_Output = "Speed," & .m_carTelemetryData(.m_header.m_playerCarIndex).m_speed
                    'Dash_2_Output = "Gear," & .m_carTelemetryData(.m_header.m_playerCarIndex).m_gear
                    'Dash_3_Output = "RPM," & .m_carTelemetryData(.m_header.m_playerCarIndex).m_engineRPM
                    'Dash_4_Output = "max RPM," & .m_carStatusData(.m_header.m_playerCarIndex).m_maxRPM
                    'Dash_5_Output = "," & .
                    'Dash_6_Output = "Fuel," & .m_carStatusData(.m_header.m_playerCarIndex).m_fuelInTank
                    'Dash_7_Output = "Fuel capacity," & .m_carStatusData(.m_header.m_playerCarIndex).m_fuelCapacity
                    'Dash_8_Output = "Flags," & .m_carStatusData(.m_header.m_playerCarIndex).m_vehicleFiaFlags
                    'Dash_9_Output = "Engine temp.," & .m_carTelemetryData(.m_header.m_playerCarIndex).m_engineTemperature
                    'Dash_10_Output = "Pit limiter," & .m_carStatusData(.m_header.m_playerCarIndex).m_pitLimiterStatus
                    'Dash_11_Output = "In pits," & .m_lapData(.m_header.m_playerCarIndex).m_pitStatus
                    'Dash_12_Output = "Track," & .m_trackId
                    'Dash_13_Output = "Team ID," & .m_participants(.m_header.m_playerCarIndex).m_teamId
                    'Dash_14_Output = "Name," & .m_participants(.m_header.m_playerCarIndex).m_name
                    'Dash_15_Output = "," & .
                    'Dash_16_Output = "," & .
                    'Dash_17_Output = "," & .
                    'Dash_18_Output = "," & .
                    'Dash_19_Output = "," & .
                    'Dash_20_Output = "," & .
                End With

            ElseIf ByteArray(3) = 4 Then
                Dim CodemastersData As New PacketParticipantsData
                Dim gch As GCHandle = GCHandle.Alloc(ByteArray, GCHandleType.Pinned)
                CodemastersData = Marshal.PtrToStructure(gch.AddrOfPinnedObject, CodemastersData.GetType)
                gch.Free()
                With CodemastersData
                    Dash_13_Output = "Team ID," & .m_participants(.m_header.m_playerCarIndex).m_teamId
                    Dash_14_Output = "Name," & .m_participants(.m_header.m_playerCarIndex).m_name
                End With

            ElseIf ByteArray(3) = 5 Then
                Dim CodemastersData As New PacketCarSetupData
                Dim gch As GCHandle = GCHandle.Alloc(ByteArray, GCHandleType.Pinned)
                CodemastersData = Marshal.PtrToStructure(gch.AddrOfPinnedObject, CodemastersData.GetType)
                gch.Free()
                With CodemastersData
                    '…
                End With

            ElseIf ByteArray(3) = 6 Then
                Dim CodemastersData As New PacketCarTelemetryData
                Dim gch As GCHandle = GCHandle.Alloc(ByteArray, GCHandleType.Pinned)
                CodemastersData = Marshal.PtrToStructure(gch.AddrOfPinnedObject, CodemastersData.GetType)
                gch.Free()
                With CodemastersData
                    Dash_1_Output = "Speed," & .m_carTelemetryData(.m_header.m_playerCarIndex).m_speed
                    Dash_2_Output = "Gear," & .m_carTelemetryData(.m_header.m_playerCarIndex).m_gear
                    Dash_3_Output = "RPM," & .m_carTelemetryData(.m_header.m_playerCarIndex).m_engineRPM
                    Dash_9_Output = "Engine temp.," & .m_carTelemetryData(.m_header.m_playerCarIndex).m_engineTemperature
                    Vibe_1_Output = "RPM," & Int(.m_carTelemetryData(.m_header.m_playerCarIndex).m_engineRPM + 0.5)
                    Vibe_2_Output = "Gear Shift," & (.m_carTelemetryData(.m_header.m_playerCarIndex).m_gear + 1).ToString
                    Vibe_7_Output = "Turbulence," & Int(.m_carTelemetryData(.m_header.m_playerCarIndex).m_speed + 0.5).ToString
                End With

            ElseIf ByteArray(3) = 7 Then
                Dim CodemastersData As New PacketCarStatusData
                Dim gch As GCHandle = GCHandle.Alloc(ByteArray, GCHandleType.Pinned)
                CodemastersData = Marshal.PtrToStructure(gch.AddrOfPinnedObject, CodemastersData.GetType)
                gch.Free()
                With CodemastersData
                    Dash_4_Output = "max RPM," & .m_carStatusData(.m_header.m_playerCarIndex).m_maxRPM
                    Dash_6_Output = "Fuel," & .m_carStatusData(.m_header.m_playerCarIndex).m_fuelInTank
                    Dash_7_Output = "Fuel capacity," & .m_carStatusData(.m_header.m_playerCarIndex).m_fuelCapacity
                    Dash_8_Output = "Flags," & .m_carStatusData(.m_header.m_playerCarIndex).m_vehicleFiaFlags
                    Dash_10_Output = "Pit limiter," & .m_carStatusData(.m_header.m_playerCarIndex).m_pitLimiterStatus
                End With
            End If
        Else
            Dim CodemastersData As New UDPPacket
            Dim gch As GCHandle = GCHandle.Alloc(ByteArray, GCHandleType.Pinned)
            CodemastersData = Marshal.PtrToStructure(gch.AddrOfPinnedObject, CodemastersData.GetType)
            gch.Free()

            With CodemastersData
                Yaw_Output = .m_yaw * radtodeg
                Roll_Output = .m_roll * radtodeg
                Pitch_Output = .m_pitch * radtodeg
                Surge_Output = .m_gforce_lon
                Sway_Output = .m_gforce_lat
                Heave_Output = .m_gforce_vert
                Extra1_Output = .m_x_local_velocity

                Dash_1_Output = "Speed," & .m_speed * 1.60934
                Dash_2_Output = "Gear," & .m_gear
                Dash_3_Output = "RPM," & .m_engineRate
                Dash_4_Output = "max RPM," & .m_max_rpm
                'Dash_5_Output = "," & 
                Dash_6_Output = "Fuel," & .m_fuel_in_tank
                Dash_7_Output = "Fuel capacity," & .m_fuel_capacity

                Dash_8_Output = "Flags," & .m_vehicleFIAFlags
                Dash_9_Output = "Engine temp.," & .m_engine_temperature
                Dash_10_Output = "Pit limiter," & .m_pit_limiter_status
                Dash_11_Output = "In pits," & .m_in_pits
                Dash_12_Output = "Track," & .m_track_number
                Dash_13_Output = "Team ID," & .m_team_info
                '    Dash_14_Output = "," & .
                '    Dash_15_Output = "," & .
                '    Dash_16_Output = "," & .
                '    Dash_17_Output = "," & .
                '    Dash_18_Output = "," & .
                '    Dash_19_Output = "," & .
                '    Dash_20_Output = "," & .

                Vibe_1_Output = "RPM," & Int(.m_engineRate + 0.5)
                Vibe_2_Output = "Gear Shift," & (.m_gear + 1).ToString
                Vibe_3_Output = "Collision L/R," & Sway_Output.ToString
                Vibe_4_Output = "Collision F/B," & Surge_Output.ToString
                Vibe_5_Output = "Road Detail," & Heave_Output.ToString
                Vibe_6_Output = "Chassis," & .m_susp_pos(0) & ":" & .m_susp_pos(1) & ":" & .m_susp_pos(2) & ":" & .m_susp_pos(3)
                Vibe_7_Output = "Turbulence," & Int(.m_speed * 1.60934 + 0.5).ToString
            End With

        End If
    End Sub

    'Used by GameManager to Patch a Game.
    Public Function PatchGame(ByVal MyPath As String, ByVal MyIp As String) As Boolean Implements IPlugin_Game.PatchGame
        'Always try to unpatch first
        UnPatch(MyPath)

        'Change as Needed
        If IO.File.Exists(MyPath & "\hardwaresettings\hardware_settings_config.xml") Then 'Backup Configuration File
            If IO.File.Exists(MyPath & "\hardwaresettings\cfg.bak") = False Then
                FileCopy(MyPath & "\hardwaresettings\hardware_settings_config.xml", MyPath & "\hardwaresettings\cfg.BAK")
            End If
        End If

        If IO.File.Exists(MyPath & "\hardwaresettings\hardware_settings_config_vr.xml") Then 'Backup Configuration File
            If IO.File.Exists(MyPath & "\hardwaresettings\cfgvr.bak") = False Then
                FileCopy(MyPath & "\hardwaresettings\hardware_settings_config_vr.xml", MyPath & "\hardwaresettings\cfgvr.BAK")
            End If
        End If

        Try
            'No VR
            If IO.File.Exists(MyPath & "\hardwaresettings\hardware_settings_config.xml") Then
                Try
                    Dim MyXML = XDocument.Load(MyPath & "\hardwaresettings\hardware_settings_config.xml")
                    MyXML.<hardware_settings_config>.<motion>.<dbox>.@enabled = "false"
                    MyXML.<hardware_settings_config>.<motion>.<udp>.@enabled = "true"
                    MyXML.<hardware_settings_config>.<motion>.<udp>.@broadcast = "false"
                    MyXML.<hardware_settings_config>.<motion>.<udp>.@ip = MyIp
                    MyXML.<hardware_settings_config>.<motion>.<udp>.@port = _Port
                    MyXML.<hardware_settings_config>.<motion>.<udp>.@sendRate = "60"
                    MyXML.<hardware_settings_config>.<motion>.<udp>.@format = "2018"
                    MyXML.Save(MyPath & "\hardwaresettings\hardware_settings_config.xml")
                Catch ex As Exception
                End Try
            End If

            'VR
            If IO.File.Exists(MyPath & "\hardwaresettings\hardware_settings_config_vr.xml") Then
                Try
                    Dim MyXML = XDocument.Load(MyPath & "\hardwaresettings\hardware_settings_config_vr.xml")
                    MyXML.<hardware_settings_config>.<motion>.<dbox>.@enabled = "false"
                    MyXML.<hardware_settings_config>.<motion>.<udp>.@enabled = "true"
                    MyXML.<hardware_settings_config>.<motion>.<udp>.@broadcast = "false"
                    MyXML.<hardware_settings_config>.<motion>.<udp>.@ip = MyIp
                    MyXML.<hardware_settings_config>.<motion>.<udp>.@port = _Port
                    MyXML.<hardware_settings_config>.<motion>.<udp>.@sendRate = "60"
                    MyXML.<hardware_settings_config>.<motion>.<udp>.@format = "2018"
                    MyXML.Save(MyPath & "\hardwaresettings\hardware_settings_config_vr.xml")
                Catch ex As Exception
                End Try
            End If

            MsgBox("Patch Installed!", MsgBoxStyle.OkOnly, "Patching info")
            Return True
        Catch ex As Exception
            MsgBox("You Must first Install and Run the Game Once Prior to Patching" + vbCrLf + "Game Not Patched!", MsgBoxStyle.OkOnly, "Patching info")
            Return False
        End Try
    End Function

    'Used by GameManager to UnPatch a Game.
    Public Sub UnPatchGame(MyPath As String) Implements IPlugin_Game.UnPatchGame
        'Change as Needed
        UnPatch(MyPath)
        MsgBox("Patch Uninstalled!", MsgBoxStyle.OkOnly, "Patching info")
    End Sub

    'Used by GameManager to UnPatch a Game.
    Private Sub UnPatch(MyPath As String)
        'No VR
        If IO.File.Exists(MyPath & "\hardwaresettings\cfg.BAK") Then 'Restore backup File
            If IO.File.Exists(MyPath & "\hardwaresettings\hardware_settings_config.xml") Then
                IO.File.Delete(MyPath & "\hardwaresettings\hardware_settings_config.xml")
            End If
            Rename(MyPath & "\hardwaresettings\cfg.BAK", MyPath & "\hardwaresettings\hardware_settings_config.xml")
        End If

        'VR
        If IO.File.Exists(MyPath & "\hardwaresettings\cfgvr.BAK") Then 'Restore backup File
            If IO.File.Exists(MyPath & "\hardwaresettings\hardware_settings_config_vr.xml") Then
                IO.File.Delete(MyPath & "\hardwaresettings\hardware_settings_config_vr.xml")
            End If
            Rename(MyPath & "\hardwaresettings\cfgvr.BAK", MyPath & "\hardwaresettings\hardware_settings_config_vr.xml")
        End If
    End Sub

    'Tells the User where to patch the game 
    Public Sub PatchPathInfo() Implements IPlugin_Game.PatchPathInfo
        'Tell the User where to patch the game 
        'MsgBox("32bit PC's - Please Select F1 2017's Installation Directory." & vbCrLf & "64bit PC's - Please Select F1 2017's Game Save Directory" & vbCrLf & "(64bit's default location is Documents\MyGames\F1 2017).", MsgBoxStyle.OkOnly, "Patching info")
        'tell them to complete a race first before patching.
        MsgBox("Before Patching, Please Complete one race in the game.", MsgBoxStyle.OkOnly, "Patching info")

        'See if they are on a 64 bit or 32 bit machine.
        'If Environment.Is64BitOperatingSystem Then
        'Is 64 bit
        MsgBox("Please Select F1 2022's Game Save Directory" & vbCrLf & "(Win10 default location is Documents\MyGames\F1 2022).", MsgBoxStyle.OkOnly, "Patching info")
        'Else
        '    'Is 32 bit
        '    MsgBox("Please Select F1 2020's Installation Directory.", MsgBoxStyle.OkOnly, "Patching info")
        'End If
    End Sub

    'Used by GameManager to Validate a Path befors Patching.
    Public Function ValidatePatchPath(MyPath As String) As Boolean Implements IPlugin_Game.ValidatePatchPath
        'insert a simple validation of the patching path - let the user know he got it right
        If IO.File.Exists(MyPath & "\hardwaresettings\hardware_settings_config.xml") = True Then
            Return True
        Else
            Return False
        End If
    End Function


    '//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
    '///                                                 PLACE EXTRA NEEDED CODE/FUNCTIONS HERE                                                     ///
    '//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
#Region "///// - EXTRA CODE/FUNCTIONS USED FOR THIS PLUGIN - /////"
    'Used by GameEngine to Process Incoming UDP Packets.
    'https://forums.codemasters.com/discussion/136948/f1-2018-udp-specification

    <StructLayout(LayoutKind.Sequential, Pack:=1)>
    Public Structure PacketHeader
        Public m_packetFormat As UInt16          ' 2018
        Public m_packetVersion As Byte           ' Version of this packet type, all start from 1
        Public m_packetId As Byte                ' Identifier for the packet type, see below
        Public m_sessionUID As UInt64            ' Unique identifier for the session
        Public m_sessionTime As Single           ' Session timestamp
        Public m_frameIdentifier As UInt32       ' Identifier for the frame the data was retrieved on
        Public m_playerCarIndex As Byte          ' Index of player's car in the array
    End Structure

    <StructLayout(LayoutKind.Sequential, Pack:=1)>
    Public Structure CarMotionData
        Public m_worldPositionX As Single        ' World space X position
        Public m_worldPositionY As Single        ' World space Y position
        Public m_worldPositionZ As Single        ' World space Z position
        Public m_worldVelocityX As Single        ' Velocity in world space X
        Public m_worldVelocityY As Single        ' Velocity in world space Y
        Public m_worldVelocityZ As Single        ' Velocity in world space Z
        Public m_worldForwardDirX As Int16       ' World space forward X direction (normalised)
        Public m_worldForwardDirY As Int16       ' World space forward Y direction (normalised)
        Public m_worldForwardDirZ As Int16       ' World space forward Z direction (normalised)
        Public m_worldRightDirX As Int16         ' World space right X direction (normalised)
        Public m_worldRightDirY As Int16         ' World space right Y direction (normalised)
        Public m_worldRightDirZ As Int16         ' World space right Z direction (normalised)
        Public m_gForceLateral As Single         ' Lateral G-Force component
        Public m_gForceLongitudinal As Single    ' Longitudinal G-Force component
        Public m_gForceVertical As Single        ' Vertical G-Force component
        Public m_yaw As Single                   ' Yaw angle in radians
        Public m_pitch As Single                 ' Pitch angle in radians
        Public m_roll As Single                  ' Roll angle in radians
    End Structure

    <StructLayout(LayoutKind.Sequential, Pack:=1)>
    Public Structure PacketMotionData
        Public m_header As PacketHeader                        ' Header
        <MarshalAs(UnmanagedType.ByValArray, SizeConst:=20)>
        Public m_carMotionData() As CarMotionData              ' Data for all cars on track
        ' Extra player car ONLY data	
        <MarshalAs(UnmanagedType.ByValArray, SizeConst:=4)>
        Public m_suspensionPosition() As Single                ' Note: All wheel arrays have the following order:
        <MarshalAs(UnmanagedType.ByValArray, SizeConst:=4)>
        Public m_suspensionVelocity() As Single                ' RL, RR, FL, FR
        <MarshalAs(UnmanagedType.ByValArray, SizeConst:=4)>
        Public m_suspensionAcceleration() As Single            ' RL, RR, FL, FR
        <MarshalAs(UnmanagedType.ByValArray, SizeConst:=4)>
        Public m_wheelSpeed() As Single                        ' Speed of each wheel
        <MarshalAs(UnmanagedType.ByValArray, SizeConst:=4)>
        Public m_wheelSlip() As Single                         ' Slip ratio for each wheel
        Public m_localVelocityX As Single                      ' Velocity in local space
        Public m_localVelocityY As Single                      ' Velocity in local space
        Public m_localVelocityZ As Single                      ' Velocity in local space
        Public m_angularVelocityX As Single                    ' Angular velocity x-component
        Public m_angularVelocityY As Single                    ' Angular velocity y-component
        Public m_angularVelocityZ As Single                    ' Angular velocity z-component
        Public m_angularAccelerationX As Single                ' Angular velocity x-component
        Public m_angularAccelerationY As Single                ' Angular velocity y-component
        Public m_angularAccelerationZ As Single                ' Angular velocity z-component
        Public m_frontWheelsAngle As Single                    ' Current front wheels angle in radians
    End Structure

    <StructLayout(LayoutKind.Sequential, Pack:=1)>
    Public Structure CarTelemetryData
        'Size: 1085 bytes
        Public m_speed As UInt16                    ' Speed Of car In kilometres per hour
        Public m_throttle As Byte                   ' Amount Of throttle applied (0 To 100)
        Public m_steer As SByte                     ' Steering (-100 (full lock left) To 100 (full lock right))
        Public m_brake As Byte                      ' Amount Of brake applied (0 To 100)
        Public m_clutch As Byte                     ' Amount Of clutch applied (0 To 100)
        Public m_gear As SByte                      ' Gear selected (1-8, N=0, R=-1)
        Public m_engineRPM As UInt16                ' Engine RPM
        Public m_drs As Byte                        ' 0 = off, 1 = On
        Public m_revLightsPercent As Byte           ' Rev lights indicator (percentage)
        <MarshalAs(UnmanagedType.ByValArray, SizeConst:=4)>
        Public m_brakesTemperature() As UInt16       ' Brakes temperature (celsius)
        <MarshalAs(UnmanagedType.ByValArray, SizeConst:=4)>
        Public m_tyresSurfaceTemperature() As UInt16 ' Tyres surface temperature (celsius)
        <MarshalAs(UnmanagedType.ByValArray, SizeConst:=4)>
        Public m_tyresInnerTemperature() As UInt16   ' Tyres inner temperature (celsius)
        Public m_engineTemperature As UInt16         ' Engine temperature (celsius)
        <MarshalAs(UnmanagedType.ByValArray, SizeConst:=4)>
        Public m_tyresPressure() As Single           ' Tyres pressure (PSI)
    End Structure

    <StructLayout(LayoutKind.Sequential, Pack:=1)>
    Public Structure PacketCarTelemetryData

        Public m_header As PacketHeader           ' Header
        <MarshalAs(UnmanagedType.ByValArray, SizeConst:=20)>
        Public m_carTelemetryData() As CarTelemetryData
        Public m_buttonStatus As UInt32           ' Bit flags specifying which buttons are being
        '                                           pressed currently - see appendices
    End Structure

    <StructLayout(LayoutKind.Sequential, Pack:=1)>
    Public Structure CarStatusData

        Public m_tractionControl As Byte          ' 0 (off) - 2 (high)
        Public m_antiLockBrakes As Byte           ' 0 (off) - 1 (On)
        Public m_fuelMix As Byte                  ' Fuel mix - 0 = lean, 1 = standard, 2 = rich, 3 = max
        Public m_frontBrakeBias As Byte           ' Front brake bias (percentage)
        Public m_pitLimiterStatus As Byte         ' Pit limiter status - 0 = off, 1 = On
        Public m_fuelInTank As Single             ' Current fuel mass
        Public m_fuelCapacity As Single           ' Fuel capacity
        Public m_maxRPM As UInt16                 ' Cars max RPM, point Of rev limiter
        Public m_idleRPM As UInt16                ' Cars idle RPM
        Public m_maxGears As Byte                 ' Maximum number Of gears
        Public m_drsAllowed As Byte               ' 0 = Not allowed, 1 = allowed, -1 = unknown
        <MarshalAs(UnmanagedType.ByValArray, SizeConst:=4)>
        Public m_tyresWear() As Byte              ' Tyre wear percentage
        Public m_tyreCompound As Byte              ' Modern - 0 = hyper soft, 1 = ultra soft
        '                                           2 = super soft, 3 = soft, 4 = medium, 5 = hard
        '                                           6 = super hard, 7 = inter, 8 = wet
        '                                           Classic - 0-6 = dry, 7-8 = wet
        <MarshalAs(UnmanagedType.ByValArray, SizeConst:=4)>
        Public m_tyresDamage() As Byte            ' Tyre damage (percentage)
        Public m_frontLeftWingDamage As Byte      ' Front left wing damage (percentage)
        Public m_frontRightWingDamage As Byte     ' Front right wing damage (percentage)
        Public m_rearWingDamage As Byte           ' Rear wing damage (percentage)
        Public m_engineDamage As Byte             ' Engine damage (percentage)
        Public m_gearBoxDamage As Byte            ' Gear box damage (percentage)
        Public m_exhaustDamage As Byte            ' Exhaust damage (percentage)
        Public m_vehicleFiaFlags As SByte         ' -1 = invalid/unknown, 0 = none, 1 = green
        '                                           2 = blue, 3 = yellow, 4 = red
        Public m_ersStoreEnergy As Single         ' ERS energy store In Joules
        Public m_ersDeployMode As Byte            ' ERS deployment mode, 0 = none, 1 = low, 2 = medium
        '                                           3 = high, 4 = overtake, 5 = hotlap
        Public m_ersHarvestedThisLapMGUK As Single ' ERS energy harvested this lap by MGU-K
        Public m_ersHarvestedThisLapMGUH As Single ' ERS energy harvested this lap by MGU-H
        Public m_ersDeployedThisLap As Single      ' ERS energy deployed this lap
    End Structure

    <StructLayout(LayoutKind.Sequential, Pack:=1)>
    Public Structure PacketCarStatusData
        Public m_header As PacketHeader         ' Header
        <MarshalAs(UnmanagedType.ByValArray, SizeConst:=20)>
        Public m_carStatusData() As CarStatusData
    End Structure

    <StructLayout(LayoutKind.Sequential, Pack:=1)>
    Public Structure ParticipantData
        Public m_aiControlled As Byte           ' Whether the vehicle Is AI (1) Or Human (0) controlled
        Public m_driverId As Byte               ' Driver id - see appendix
        Public m_teamId As Byte                 ' Team id - see appendix
        Public m_raceNumber As Byte             ' Race number Of the car
        Public m_nationality As Byte            ' Nationality Of the driver
        <MarshalAs(UnmanagedType.ByValTStr, SizeConst:=48)>
        Public m_name As String                 ' Name of participant in UTF-8 format – null terminated
        '                                         Will be truncated with … (U+2026) if too long
    End Structure

    <StructLayout(LayoutKind.Sequential, Pack:=1)>
    Public Structure PacketParticipantsData

        Public m_header As PacketHeader            ' Header
        Public m_numCars As Byte           ' Number Of cars In the data
        <MarshalAs(UnmanagedType.ByValArray, SizeConst:=20)>
        Public m_participants() As ParticipantData
    End Structure

    <StructLayout(LayoutKind.Sequential, Pack:=1)>
    Public Structure MarshalZone

        Public m_zoneStart As Single   ' Fraction (0..1) Of way through the lap the marshal zone starts
        Public m_zoneFlag As SByte    ' -1 = invalid/unknown, 0 = none, 1 = green, 2 = blue, 3 = yellow, 4 = red
    End Structure

    <StructLayout(LayoutKind.Sequential, Pack:=1)>
    Public Structure PacketSessionData

        Public m_header As PacketHeader           ' Header
        Public m_weather As Byte                  ' Weather - 0 = clear, 1 = light cloud, 2 = overcast
        '                                           3 = light rain, 4 = heavy rain, 5 = storm
        Public m_trackTemperature As SByte         ' Track temp. In degrees celsius
        Public m_airTemperature As SByte           ' Air temp. In degrees celsius
        Public m_totalLaps As Byte                 ' Total number Of laps In this race
        Public m_trackLength As UInt16             ' Track length In metres
        Public m_sessionType As Byte               ' 0 = unknown, 1 = P1, 2 = P2, 3 = P3, 4 = Short P
        '                                            5 = Q1, 6 = Q2, 7 = Q3, 8 = Short Q, 9 = OSQ
        '                                            10 = R, 11 = R2, 12 = Time Trial
        Public m_trackId As SByte                  ' -1 For unknown, 0-21 For tracks, see appendix
        Public m_era As Byte                       ' Era, 0 = modern, 1 = classic
        Public m_sessionTimeLeft As UInt16         ' Time left In session In seconds
        Public m_sessionDuration As UInt16         ' Session duration In seconds
        Public m_pitSpeedLimit As Byte             ' Pit speed limit In kilometres per hour
        Public m_gamePaused As Byte                ' Whether the game Is paused
        Public m_isSpectating As Byte              ' Whether the player Is spectating
        Public m_spectatorCarIndex As Byte         ' Index Of the car being spectated
        Public m_sliProNativeSupport As Byte       ' SLI Pro support, 0 = inactive, 1 = active
        Public m_numMarshalZones As Byte           ' Number Of marshal zones To follow
        <MarshalAs(UnmanagedType.ByValArray, SizeConst:=21)>
        Public m_marshalZones() As MarshalZone     ' List Of marshal zones – max 21
        Public m_safetyCarStatus As Byte           ' 0 = no safety car, 1 = full safety car
        '                                            2 = virtual safety car
        Public m_networkGame As Byte               ' 0 = offline, 1 = online
    End Structure

    <StructLayout(LayoutKind.Sequential, Pack:=1)>
    Public Structure LapData

        Public m_lastLapTime As Single           ' Last lap time In seconds
        Public m_currentLapTime As Single        ' Current time around the lap In seconds
        Public m_bestLapTime As Single           ' Best lap time Of the session In seconds
        Public m_sector1Time As Single           ' Sector 1 time In seconds
        Public m_sector2Time As Single           ' Sector 2 time In seconds
        Public m_lapDistance As Single           ' Distance vehicle Is around current lap In metres – could
        '                                          be negative if line hasn't been crossed yet
        Public m_totalDistance As Single         ' Total distance travelled In session In metres – could
        '                                          be negative if line hasn't been crossed yet
        Public m_safetyCarDelta As Single        ' Delta In seconds For safety car
        Public m_carPosition As Byte             ' Car race position
        Public m_currentLapNum As Byte           ' Current lap number
        Public m_pitStatus As Byte               ' 0 = none, 1 = pitting, 2 = In pit area
        Public m_sector As Byte                  ' 0 = sector1, 1 = sector2, 2 = sector3
        Public m_currentLapInvalid As Byte       ' Current lap invalid - 0 = valid, 1 = invalid
        Public m_penalties As Byte               ' Accumulated time penalties In seconds To be added
        Public m_gridPosition As Byte            ' Grid position the vehicle started the race In
        Public m_driverStatus As Byte            ' Status Of driver - 0 = In garage, 1 = flying lap
        '                                          2 = in lap, 3 = out lap, 4 = on track
        Public m_resultStatus As Byte            ' Result status - 0 = invalid, 1 = inactive, 2 = active
        '                                          3 = finished, 4 = disqualified, 5 = Not classified
        '                                          6 = retired
    End Structure

    <StructLayout(LayoutKind.Sequential, Pack:=1)>
    Public Structure PacketLapData

        Public m_header As PacketHeader           ' Header
        <MarshalAs(UnmanagedType.ByValArray, SizeConst:=20)>
        Public m_lapData() As LapData             ' Lap data For all cars On track
    End Structure

    <StructLayout(LayoutKind.Sequential, Pack:=1)>
    Public Structure PacketEventData

        Public m_header As PacketHeader           ' Header
        <MarshalAs(UnmanagedType.ByValArray, SizeConst:=4)>
        Public m_eventStringCode() As Byte        ' Event String code, see above
    End Structure

    <StructLayout(LayoutKind.Sequential, Pack:=1)>
    Public Structure CarSetupData

        Public m_frontWing As Byte                ' Front wing aero
        Public m_rearWing As Byte                 ' Rear wing aero
        Public m_onThrottle As Byte               ' Differential adjustment On throttle (percentage)
        Public m_offThrottle As Byte              ' Differential adjustment off throttle (percentage)
        Public m_frontCamber As Single            ' Front camber angle (suspension geometry)
        Public m_rearCamber As Single             ' Rear camber angle (suspension geometry)
        Public m_frontToe As Single               ' Front toe angle (suspension geometry)
        Public m_rearToe As Single                ' Rear toe angle (suspension geometry)
        Public m_frontSuspension As Byte          ' Front suspension
        Public m_rearSuspension As Byte           ' Rear suspension
        Public m_frontAntiRollBar As Byte         ' Front anti-roll bar
        Public m_rearAntiRollBar As Byte          ' Front anti-roll bar
        Public m_frontSuspensionHeight As Byte    ' Front ride height
        Public m_rearSuspensionHeight As Byte     ' Rear ride height
        Public m_brakePressure As Byte            ' Brake pressure (percentage)
        Public m_brakeBias As Byte                ' Brake bias (percentage)
        Public m_frontTyrePressure As Single      ' Front tyre pressure (PSI)
        Public m_rearTyrePressure As Single       ' Rear tyre pressure (PSI)
        Public m_ballast As Byte                  ' Ballast
        Public m_fuelLoad As Single               ' Fuel load
    End Structure

    <StructLayout(LayoutKind.Sequential, Pack:=1)>
    Public Structure PacketCarSetupData

        Public m_header As PacketHeader           ' Header
        <MarshalAs(UnmanagedType.ByValArray, SizeConst:=20)>
        Public m_carSetups() As CarSetupData
    End Structure


    ' * * * * *   L E G A C Y   F O R M A T   * * * * *
    <StructLayout(LayoutKind.Sequential, Pack:=1)>
    Public Structure UDPPacket
        Public m_time As Single
        Public m_lapTime As Single
        Public m_lapDistance As Single
        Public m_totalDistance As Single
        Public m_x As Single     ' World space position
        Public m_y As Single     ' World space position
        Public m_z As Single     ' World space position
        Public m_speed As Single     ' Speed of car in MPH
        Public m_xv As Single    ' Velocity in world space
        Public m_yv As Single    ' Velocity in world space
        Public m_zv As Single    ' Velocity in world space
        Public m_xr As Single    ' World space right direction
        Public m_yr As Single    ' World space right direction
        Public m_zr As Single    ' World space right direction
        Public m_xd As Single    ' World space forward direction
        Public m_yd As Single    ' World space forward direction
        Public m_zd As Single    ' World space forward direction
        <MarshalAs(UnmanagedType.ByValArray, SizeConst:=4)>
        Public m_susp_pos() As Single    ' Note: All wheel arrays have the order:
        <MarshalAs(UnmanagedType.ByValArray, SizeConst:=4)>
        Public m_susp_vel() As Single    ' RL, RR, FL, FR
        <MarshalAs(UnmanagedType.ByValArray, SizeConst:=4)>
        Public m_wheel_speed() As Single
        Public m_throttle As Single
        Public m_steer As Single
        Public m_brake As Single
        Public m_clutch As Single
        Public m_gear As Single
        Public m_gforce_lat As Single
        Public m_gforce_lon As Single
        Public m_lap As Single
        Public m_engineRate As Single
        Public m_sli_pro_native_support As Single    ' SLI Pro support
        Public m_car_position As Single  ' car race position
        Public m_kers_level As Single    ' kers energy left
        Public m_kers_max_level As Single    ' kers maximum energy
        Public m_drs As Single   ' 0 = off, 1 = on
        Public m_traction_control As Single  ' 0 (off) - 2 (high)
        Public m_anti_lock_brakes As Single  ' 0 (off) - 1 (on)
        Public m_fuel_in_tank As Single  ' current fuel mass
        Public m_fuel_capacity As Single     ' fuel capacity
        Public m_in_pits As Single   ' 0 = none, 1 = pitting, 2 = in pit area
        Public m_sector As Single    ' 0 = sector1, 1 = sector2, 2 = sector3
        Public m_sector1_time As Single  ' time of sector1 (or 0)
        Public m_sector2_time As Single  ' time of sector2 (or 0)
        <MarshalAs(UnmanagedType.ByValArray, SizeConst:=4)>
        Public m_brakes_temp() As Single     ' brakes temperature (centigrade)
        <MarshalAs(UnmanagedType.ByValArray, SizeConst:=4)>
        Public m_tyres_pressure() As Single  ' tyres pressure PSI
        Public m_team_info As Single     ' team ID 
        Public m_total_laps As Single    ' total number of laps in this race
        Public m_track_size As Single    ' track size meters
        Public m_last_lap_time As Single     ' last lap time
        Public m_max_rpm As Single   ' cars max RPM, at which point the rev limiter will kick in
        Public m_idle_rpm As Single  ' cars idle RPM
        Public m_max_gears As Single     ' maximum number of gears
        Public m_sessionType As Single   ' 0 = unknown, 1 = practice, 2 = qualifying, 3 = race
        Public m_drsAllowed As Single    ' 0 = not allowed, 1 = allowed, -1 = invalid / unknown
        Public m_track_number As Single  ' -1 for unknown, 0-21 for tracks
        Public m_vehicleFIAFlags As Single   ' -1 = invalid/unknown, 0 = none, 1 = green, 2 = blue, 3 = yellow, 4 = red
        Public m_era As Single                       ' era, 2017 (modern) or 1980 (classic)
        Public m_engine_temperature As Single      ' engine temperature (centigrade)
        Public m_gforce_vert As Single   ' vertical g-force component
        Public m_ang_vel_x As Single     ' angular velocity x-component
        Public m_ang_vel_y As Single     ' angular velocity y-component
        Public m_ang_vel_z As Single     ' angular velocity z-component
        <MarshalAs(UnmanagedType.ByValArray, SizeConst:=4)>
        Public m_tyres_temperature() As Byte   ' tyres temperature (centigrade)
        <MarshalAs(UnmanagedType.ByValArray, SizeConst:=4)>
        Public m_tyres_wear() As Byte    ' tyre wear percentage
        Public m_tyre_compound As Byte   ' compound of tyre – 0 = ultra soft, 1 = super soft, 2 = soft, 3 = medium, 4 = hard, 5 = inter, 6 = wet
        Public m_front_brake_bias As Byte            ' front brake bias (percentage)
        Public m_fuel_mix As Byte                    ' fuel mix - 0 = lean, 1 = standard, 2 = rich, 3 = max
        Public m_currentLapInvalid As Byte       ' current lap invalid - 0 = valid, 1 = invalid
        <MarshalAs(UnmanagedType.ByValArray, SizeConst:=4)>
        Public m_tyres_damage() As Byte  ' tyre damage (percentage)
        Public m_front_left_wing_damage As Byte  ' front left wing damage (percentage)
        Public m_front_right_wing_damage As Byte     ' front right wing damage (percentage)
        Public m_rear_wing_damage As Byte    ' rear wing damage (percentage)
        Public m_engine_damage As Byte   ' engine damage (percentage)
        Public m_gear_box_damage As Byte     ' gear box damage (percentage)
        Public m_exhaust_damage As Byte  ' exhaust damage (percentage)
        Public m_pit_limiter_status As Byte  ' pit limiter status – 0 = off, 1 = on
        Public m_pit_speed_limit As Byte     ' pit speed limit in mph
        Public m_session_time_left As Single      ' NEW: time left in session in seconds 
        Public m_rev_lights_percent As Byte   ' NEW: rev lights indicator (percentage)
        Public m_is_spectating As Byte    ' NEW: whether the player is spectating
        Public m_spectator_car_index As Byte      ' NEW: index of the car being spectated
        Public m_num_cars As Byte                  ' number of cars in data
        Public m_player_car_index As Byte            ' index of player's car in the array
        <MarshalAs(UnmanagedType.ByValArray, SizeConst:=20)>
        Public m_car_data() As CarUDPData      ' data for all cars on track
        Public m_yaw As Single   ' New (v1.8)
        Public m_pitch As Single  ' New (v1.8)
        Public m_roll As Single  ' New (v1.8)
        Public m_x_local_velocity As Single          ' New (v1.8) Velocity In local space
        Public m_y_local_velocity As Single          ' New (v1.8) Velocity In local space
        Public m_z_local_velocity As Single          ' New (v1.8) Velocity In local space
        <MarshalAs(UnmanagedType.ByValArray, SizeConst:=4)>
        Public m_susp_acceleration() As Single       ' New (v1.8) RL, RR, FL, FR
        Public m_ang_acc_x As Single                 ' New (v1.8) angular acceleration x-component
        Public m_ang_acc_y As Single                 ' New (v1.8) angular acceleration y-component
        Public m_ang_acc_z As Single                 ' New (v1.8) angular acceleration z-component

    End Structure

    <StructLayout(LayoutKind.Sequential, Pack:=1)>
    Public Structure CarUDPData
        <MarshalAs(UnmanagedType.ByValArray, SizeConst:=3)>
        Public m_worldPosition() As Single  ' world co-ordinates of vehicle
        Public m_lastLapTime As Single
        Public m_currentLapTime As Single
        Public m_bestLapTime As Single
        Public m_sector1Time As Single
        Public m_sector2Time As Single
        Public m_lapDistance As Single
        Public m_driverId As Byte
        Public m_teamId As Byte
        Public m_carPosition As Byte       ' UPDATED: track positions of vehicle
        Public m_currentLapNum As Byte
        Public m_tyreCompound As Byte      ' compound of tyre – 0 = ultra soft, 1 = super soft, 2 = soft, 3 = medium, 4 = hard, 5 = inter, 6 = wet
        Public m_inPits As Byte            ' 0 = none, 1 = pitting, 2 = in pit area
        Public m_sector As Byte            ' 0 = sector1, 1 = sector2, 2 = sector3
        Public m_currentLapInvalid As Byte ' current lap invalid - 0 = valid, 1 = invalid
        Public m_penalties As Byte         ' NEW: accumulated time penalties in seconds to be added
    End Structure

    ' * * * * *   L E G A C Y   F O R M A T   * * * * *
    Enum Track
        Melbourne
        Sepang
        Shanghai
        Sakhir_Bahrain
        Catalunya
        Monaco
        Montreal
        Silverstone
        Hockenheim
        Hungaroring
        Spa
        Monza
        Singapore
        Suzuka
        Abu_Dhabi
        Texas
        Brazil
        Austria
        Sochi
        Mexico
        Baku_Azerbaijan
        Sakhir_Short
        Silverstone_Short
        Texas_Short
        Suzuka_Short
    End Enum

    Enum Team
        Redbull
        Ferrari
        McLaren
        Renault
        Mercedes
        Sauber
        Force_India
        Williams
        Toro_Rosso
        Haas = 11
    End Enum

    Enum Classic_Team
        Williams_1992
        McLaren_1988
        McLaren_2008
        Ferrari_2004
        Ferrari_1995
        Ferrari_2007
        McLaren_1998
        Williams_1996
        Renault_2006
        Ferrari_2002 = 10
        Redbull_2010
        McLaren_1991
    End Enum

    Enum Driver
        Sebastian_Vettel
        Daniil_Kvyat
        Fernando_Alonso
        Felipe_Massa
        Sergio_Perez = 5
        Kimi_Räikkönen
        Romain_Grosjean
        Lewis_Hamilton = 9
        Nico_Hulkenburg
        Kevin_Magnussen = 14
        Valtteri_Bottas
        Daniel_Ricciardo
        Marcus_Ericsson = 18
        Jolyon_Palmer = 20
        Max_Verstappen = 22
        Carlos_Sainz_Jr
        Pascal_Wehrlein = 31
        Esteban_Ocon = 33
        Stoffel_Vandoorne
        Lance_Stroll
    End Enum

    Enum Classic_Driver
        Klimek_Michalski
        Martin_Giles
        Igor_Correia
        Sophie_Levasseur
        Alain_Forest
        Santiago_Moreno
        Esto_Saari
        Peter_Belousov
        Lars_Kaufmann
        Yasar_Atiyeh
        Howard_Clarke
        Marie_Laursen = 14
        Benjamin_Coppens
        Alex_Murray
        Callisto_Calabresi = 18
        Jay_Letourneau = 20
        Naota_Izum = 22
        Arron_Barnes
        Jonas_Schiffer
        Flavio_Nieves = 31
        Noah_Visser
        Gert_Waldmuller
        Julian_Quesada
        Lucas_Roth = 68
    End Enum
#End Region


    '//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
    '///                                                        DO NOT EDIT BELOW HERE!!!                                                           ///
    '//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
#Region "///// BUILT IN METHODS FOR SIMTOOLS - DO NOT CHANGE /////"
    'Output Dash Vars
    Public Dash_1_Output As String = ""
    Public Dash_2_Output As String = ""
    Public Dash_3_Output As String = ""
    Public Dash_4_Output As String = ""
    Public Dash_5_Output As String = ""
    Public Dash_6_Output As String = ""
    Public Dash_7_Output As String = ""
    Public Dash_8_Output As String = ""
    Public Dash_9_Output As String = ""
    Public Dash_10_Output As String = ""
    Public Dash_11_Output As String = ""
    Public Dash_12_Output As String = ""
    Public Dash_13_Output As String = ""
    Public Dash_14_Output As String = ""
    Public Dash_15_Output As String = ""
    Public Dash_16_Output As String = ""
    Public Dash_17_Output As String = ""
    Public Dash_18_Output As String = ""
    Public Dash_19_Output As String = ""
    Public Dash_20_Output As String = ""

    'Output Vars
    Public Roll_Output As Double = 0
    Public Pitch_Output As Double = 0
    Public Heave_Output As Double = 0
    Public Yaw_Output As Double = 0
    Public Sway_Output As Double = 0
    Public Surge_Output As Double = 0
    Public Extra1_Output As Double = 0
    Public Extra2_Output As Double = 0
    Public Extra3_Output As Double = 0

    'MemHook Vars
    Public Roll_MemHook As Double = 0
    Public Pitch_MemHook As Double = 0
    Public Heave_MemHook As Double = 0
    Public Yaw_MemHook As Double = 0
    Public Sway_MemHook As Double = 0
    Public Surge_MemHook As Double = 0
    Public Extra1_MemHook As Double = 0
    Public Extra2_MemHook As Double = 0
    Public Extra3_MemHook As Double = 0

    'MemMap Vars
    Public Roll_MemMap As Double = 0
    Public Pitch_MemMap As Double = 0
    Public Heave_MemMap As Double = 0
    Public Yaw_MemMap As Double = 0
    Public Sway_MemMap As Double = 0
    Public Surge_MemMap As Double = 0
    Public Extra1_MemMap As Double = 0
    Public Extra2_MemMap As Double = 0
    Public Extra3_MemMap As Double = 0

    'GameVibe Vars
    Public Vibe_1_Output As String = ""
    Public Vibe_2_Output As String = ""
    Public Vibe_3_Output As String = ""
    Public Vibe_4_Output As String = ""
    Public Vibe_5_Output As String = ""
    Public Vibe_6_Output As String = ""
    Public Vibe_7_Output As String = ""
    Public Vibe_8_Output As String = ""
    Public Vibe_9_Output As String = ""

    Public Function Get_PluginVersion() As String Implements IPlugin_Game.Get_PluginVersion
        Return System.Reflection.Assembly.GetExecutingAssembly.GetName.Version.ToString
    End Function

    Public Function GetDOFsUsed() As String Implements IPlugin_Game.GetDOFsUsed
        'Return DOF's Used (Roll,Pitch,Heave,Yaw,Sway,Surge)
        Return (_DOF_Support_Roll.ToString & "," & _DOF_Support_Pitch.ToString & "," & _DOF_Support_Heave.ToString & "," & _DOF_Support_Yaw.ToString & "," & _DOF_Support_Sway.ToString & "," & _DOF_Support_Surge.ToString & "," & _DOF_Support_Extra1.ToString & "," & _DOF_Support_Extra2.ToString & "," & _DOF_Support_Extra3.ToString)
    End Function

    Public Sub ResetDOFVars() Implements IPlugin_Game.ResetDOFVars
        Roll_Output = 0
        Pitch_Output = 0
        Heave_Output = 0
        Yaw_Output = 0
        Sway_Output = 0
        Surge_Output = 0
        Extra1_Output = 0
        Extra2_Output = 0
        Extra3_Output = 0
    End Sub

    Public Sub ResetMapVars() Implements IPlugin_Game.ResetMapVars
        Roll_MemMap = 0
        Pitch_MemMap = 0
        Heave_MemMap = 0
        Yaw_MemMap = 0
        Sway_MemMap = 0
        Surge_MemMap = 0
        Extra1_MemMap = 0
        Extra2_MemMap = 0
        Extra3_MemMap = 0
    End Sub

    Public Sub ResetHookVars() Implements IPlugin_Game.ResetHookVars
        Roll_MemHook = 0
        Pitch_MemHook = 0
        Heave_MemHook = 0
        Yaw_MemHook = 0
        Sway_MemHook = 0
        Surge_MemHook = 0
        Extra1_MemHook = 0
        Extra2_MemHook = 0
        Extra3_MemHook = 0
    End Sub

    Public Sub ResetDashVars() Implements IPlugin_Game.ResetDashVars
        Dash_1_Output = ""
        Dash_2_Output = ""
        Dash_3_Output = ""
        Dash_4_Output = ""
        Dash_5_Output = ""
        Dash_6_Output = ""
        Dash_7_Output = ""
        Dash_8_Output = ""
        Dash_9_Output = ""
        Dash_10_Output = ""
        Dash_11_Output = ""
        Dash_12_Output = ""
        Dash_13_Output = ""
        Dash_14_Output = ""
        Dash_15_Output = ""
        Dash_16_Output = ""
        Dash_17_Output = ""
        Dash_18_Output = ""
        Dash_19_Output = ""
        Dash_20_Output = ""
    End Sub

    Public Sub ResetVibeVars() Implements IPlugin_Game.ResetVibeVars
        Vibe_1_Output = ""
        Vibe_2_Output = ""
        Vibe_3_Output = ""
        Vibe_4_Output = ""
        Vibe_5_Output = ""
        Vibe_6_Output = ""
        Vibe_7_Output = ""
        Vibe_8_Output = ""
        Vibe_9_Output = ""
    End Sub

    Public ReadOnly Property PluginAuthorsName() As String Implements IPlugin_Game.PluginAuthorsName
        Get
            Return _PluginAuthorsName
        End Get
    End Property

    Public ReadOnly Property Name() As String Implements IPlugin_Game.GameName
        Get
            Return _GameName
        End Get
    End Property

    Public ReadOnly Property ProcessName() As String Implements IPlugin_Game.ProcessName
        Get
            Return _ProcessName
        End Get
    End Property

    Public ReadOnly Property Port() As String Implements IPlugin_Game.Port
        Get
            Return _Port
        End Get
    End Property

    Public ReadOnly Property Enable_MemoryMap() As Boolean Implements IPlugin_Game.Enable_MemoryMap
        Get
            Return _Enable_MemoryMap
        End Get
    End Property

    Public ReadOnly Property Enable_MemoryHook() As Boolean Implements IPlugin_Game.Enable_MemoryHook
        Get
            Return _Enable_MemoryHook
        End Get
    End Property

    Public ReadOnly Property RequiresPatchingPath() As Boolean Implements IPlugin_Game.RequiresPatchingPath
        Get
            Return _RequiresPatchingPath
        End Get
    End Property

    Public ReadOnly Property RequiresSecondCheck() As Boolean Implements IPlugin_Game.RequiresSecondCheck
        Get
            Return _RequiresSecondCheck
        End Get
    End Property

    Public ReadOnly Property Enable_DashBoard() As Boolean Implements IPlugin_Game.Enable_DashBoard
        Get
            Return _Enable_DashBoard
        End Get
    End Property

    Public ReadOnly Property Enable_GameVibe() As Boolean Implements IPlugin_Game.Enable_GameVibe
        Get
            Return _Enable_GameVibe
        End Get
    End Property

    Public Function Get_RollOutput() As Double Implements IPlugin_Game.Get_RollOutput
        Return Roll_Output
    End Function

    Public Function Get_PitchOutput() As Double Implements IPlugin_Game.Get_PitchOutput
        Return Pitch_Output
    End Function

    Public Function Get_HeaveOutput() As Double Implements IPlugin_Game.Get_HeaveOutput
        Return Heave_Output
    End Function

    Public Function Get_YawOutput() As Double Implements IPlugin_Game.Get_YawOutput
        Return Yaw_Output
    End Function

    Public Function Get_SwayOutput() As Double Implements IPlugin_Game.Get_SwayOutput
        Return Sway_Output
    End Function

    Public Function Get_SurgeOutput() As Double Implements IPlugin_Game.Get_SurgeOutput
        Return Surge_Output
    End Function

    Public Function Get_Extra1Output() As Double Implements IPlugin_Game.Get_Extra1Output
        Return Extra1_Output
    End Function

    Public Function Get_Extra2Output() As Double Implements IPlugin_Game.Get_Extra2Output
        Return Extra2_Output
    End Function

    Public Function Get_Extra3Output() As Double Implements IPlugin_Game.Get_Extra3Output
        Return Extra3_Output
    End Function

    Public Function Get_Dash_1_Output() As String Implements IPlugin_Game.Get_Dash1_Output
        Return Dash_1_Output
    End Function

    Public Function Get_Dash_2_Output() As String Implements IPlugin_Game.Get_Dash2_Output
        Return Dash_2_Output
    End Function

    Public Function Get_Dash_3_Output() As String Implements IPlugin_Game.Get_Dash3_Output
        Return Dash_3_Output
    End Function

    Public Function Get_Dash_4_Output() As String Implements IPlugin_Game.Get_Dash4_Output
        Return Dash_4_Output
    End Function

    Public Function Get_Dash_5_Output() As String Implements IPlugin_Game.Get_Dash5_Output
        Return Dash_5_Output
    End Function

    Public Function Get_Dash_6_Output() As String Implements IPlugin_Game.Get_Dash6_Output
        Return Dash_6_Output
    End Function

    Public Function Get_Dash_7_Output() As String Implements IPlugin_Game.Get_Dash7_Output
        Return Dash_7_Output
    End Function

    Public Function Get_Dash_8_Output() As String Implements IPlugin_Game.Get_Dash8_Output
        Return Dash_8_Output
    End Function

    Public Function Get_Dash_9_Output() As String Implements IPlugin_Game.Get_Dash9_Output
        Return Dash_9_Output
    End Function

    Public Function Get_Dash_10_Output() As String Implements IPlugin_Game.Get_Dash10_Output
        Return Dash_10_Output
    End Function

    Public Function Get_Dash_11_Output() As String Implements IPlugin_Game.Get_Dash11_Output
        Return Dash_11_Output
    End Function

    Public Function Get_Dash_12_Output() As String Implements IPlugin_Game.Get_Dash12_Output
        Return Dash_12_Output
    End Function

    Public Function Get_Dash_13_Output() As String Implements IPlugin_Game.Get_Dash13_Output
        Return Dash_13_Output
    End Function

    Public Function Get_Dash_14_Output() As String Implements IPlugin_Game.Get_Dash14_Output
        Return Dash_14_Output
    End Function

    Public Function Get_Dash_15_Output() As String Implements IPlugin_Game.Get_Dash15_Output
        Return Dash_15_Output
    End Function

    Public Function Get_Dash_16_Output() As String Implements IPlugin_Game.Get_Dash16_Output
        Return Dash_16_Output
    End Function

    Public Function Get_Dash_17_Output() As String Implements IPlugin_Game.Get_Dash17_Output
        Return Dash_17_Output
    End Function

    Public Function Get_Dash_18_Output() As String Implements IPlugin_Game.Get_Dash18_Output
        Return Dash_18_Output
    End Function

    Public Function Get_Dash_19_Output() As String Implements IPlugin_Game.Get_Dash19_Output
        Return Dash_19_Output
    End Function

    Public Function Get_Dash_20_Output() As String Implements IPlugin_Game.Get_Dash20_Output
        Return Dash_20_Output
    End Function

    Public Function Get_RollMemHook() As Double Implements IPlugin_Game.Get_RollMemHook
        Return Roll_MemHook
    End Function

    Public Function Get_PitchMemHook() As Double Implements IPlugin_Game.Get_PitchMemHook
        Return Pitch_MemHook
    End Function

    Public Function Get_HeaveMemHook() As Double Implements IPlugin_Game.Get_HeaveMemHook
        Return Heave_MemHook
    End Function

    Public Function Get_YawMemHook() As Double Implements IPlugin_Game.Get_YawMemHook
        Return Yaw_MemHook
    End Function

    Public Function Get_SwayMemHook() As Double Implements IPlugin_Game.Get_SwayMemHook
        Return Sway_MemHook
    End Function

    Public Function Get_SurgeMemHook() As Double Implements IPlugin_Game.Get_SurgeMemHook
        Return Surge_MemHook
    End Function

    Public Function Get_Extra1MemHook() As Double Implements IPlugin_Game.Get_Extra1MemHook
        Return Extra1_MemHook
    End Function

    Public Function Get_Extra2MemHook() As Double Implements IPlugin_Game.Get_Extra2MemHook
        Return Extra2_MemHook
    End Function

    Public Function Get_Extra3MemHook() As Double Implements IPlugin_Game.Get_Extra3MemHook
        Return Extra3_MemHook
    End Function

    Public Function Get_RollMemMap() As Double Implements IPlugin_Game.Get_RollMemMap
        Return Roll_MemMap
    End Function

    Public Function Get_PitchMemMap() As Double Implements IPlugin_Game.Get_PitchMemMap
        Return Pitch_MemMap
    End Function

    Public Function Get_HeaveMemMap() As Double Implements IPlugin_Game.Get_HeaveMemMap
        Return Heave_MemMap
    End Function

    Public Function Get_YawMemMap() As Double Implements IPlugin_Game.Get_YawMemMap
        Return Yaw_MemMap
    End Function

    Public Function Get_SwayMemMap() As Double Implements IPlugin_Game.Get_SwayMemMap
        Return Sway_MemMap
    End Function

    Public Function Get_SurgeMemMap() As Double Implements IPlugin_Game.Get_SurgeMemMap
        Return Surge_MemMap
    End Function

    Public Function Get_Extra1MemMap() As Double Implements IPlugin_Game.Get_Extra1MemMap
        Return Extra1_MemMap
    End Function

    Public Function Get_Extra2MemMap() As Double Implements IPlugin_Game.Get_Extra2MemMap
        Return Extra2_MemMap
    End Function

    Public Function Get_Extra3MemMap() As Double Implements IPlugin_Game.Get_Extra3MemMap
        Return Extra3_MemMap
    End Function

    Public Function Get_Vibe1_Output() As String Implements IPlugin_Game.Get_Vibe1_Output
        Return Vibe_1_Output
    End Function

    Public Function Get_Vibe2_Output() As String Implements IPlugin_Game.Get_Vibe2_Output
        Return Vibe_2_Output
    End Function

    Public Function Get_Vibe3_Output() As String Implements IPlugin_Game.Get_Vibe3_Output
        Return Vibe_3_Output
    End Function

    Public Function Get_Vibe4_Output() As String Implements IPlugin_Game.Get_Vibe4_Output
        Return Vibe_4_Output
    End Function

    Public Function Get_Vibe5_Output() As String Implements IPlugin_Game.Get_Vibe5_Output
        Return Vibe_5_Output
    End Function

    Public Function Get_Vibe6_Output() As String Implements IPlugin_Game.Get_Vibe6_Output
        Return Vibe_6_Output
    End Function

    Public Function Get_Vibe7_Output() As String Implements IPlugin_Game.Get_Vibe7_Output
        Return Vibe_7_Output
    End Function

    Public Function Get_Vibe8_Output() As String Implements IPlugin_Game.Get_Vibe8_Output
        Return Vibe_8_Output
    End Function

    Public Function Get_Vibe9_Output() As String Implements IPlugin_Game.Get_Vibe9_Output
        Return Vibe_9_Output
    End Function

    Public ReadOnly Property PluginOptions() As String Implements IPlugin_Game.PluginOptions
        Get
            Return _PluginOptions
        End Get
    End Property
#End Region
End Class